@@ -71,12 +71,22 @@ static functor_t FUNCTOR_at1;
 static functor_t FUNCTOR_eval1;
 
 static int py_initialize_done = FALSE;
+static int py_finalizing = FALSE;
+static int py_gil_thread = 0;	/* Prolog thread that owns the GIL */
+
+typedef struct
+{ PyGILState_STATE gil;
+  int use_gil;
+} py_gil_state;
 
 static PyObject *check_error(PyObject *obj);
 static int py_unify(term_t t, PyObject *obj, int flags);
 static int py_from_prolog(term_t t, PyObject **obj);
-static void py_yield(void);
-static void py_resume(void);
+static void py_yield_first(void);
+static int py_gil_ensure(py_gil_state *state);
+static void py_gil_release(py_gil_state state);
+static void py_yield(py_gil_state state);
+static void py_resume(py_gil_state *state);
 static PyObject *py_record(term_t t);
 static int py_unify_record(term_t t, PyObject *rec);
 static int py_unify_constant(term_t t, atom_t c);
@@ -84,6 +94,7 @@ static int py_is_record(PyObject *rec);
 static PyObject *py_free_record(PyObject *rec);
 static int unchain(term_t call, PyObject **py_target);
 static PyObject *py_eval(PyObject *obj, term_t func);
+static foreign_t py_finalize(void);
 
 /* - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
 On Windows,  `python3.dll` provides the  stable C API that  relates to
@@ -267,6 +278,18 @@ compare_python_object(atom_t a, atom_t b)
 	 );
 }
 
+/* Write a Python  object reference as <py_<Class>>(Ref).   To get the
+ * class however, we need the  GIL.  Unfortunately, this may deadlock.
+ * Assume
+ *
+ *   - We printing an object ref to stream S.  We hold stream S and
+ *     try to get the GIL.
+ *   - Some other thread holds the GIL and tries to write to S.
+ *
+ * To resolve this we'd need a PyGILState_TryEnsure(), but this does
+ * not seem to exists.
+ */
+
 static int
 write_python_object(IOSTREAM *s, atom_t symbol, int flags)
 { PyObject *obj = PL_blob_data(symbol, NULL, NULL);
@@ -275,12 +298,18 @@ write_python_object(IOSTREAM *s, atom_t symbol, int flags)
   { PyObject *cls = NULL;
     PyObject *cname = NULL;
     const char *name;
+    py_gil_state state;
 
-    if ( (cls=PyObject_GetAttrString(obj, "__class__")) &&
-	 (cname=PyObject_GetAttrString(cls, "__name__")) )
-      name = PyUnicode_AsUTF8AndSize(cname, NULL);
-    else
-      name = "noclass";
+    if ( py_gil_thread == 0 && py_gil_ensure(&state) )
+    { if ( (cls=PyObject_GetAttrString(obj, "__class__")) &&
+	   (cname=PyObject_GetAttrString(cls, "__name__")) )
+	name = PyUnicode_AsUTF8AndSize(cname, NULL);
+      else
+	name = "noclass";
+      py_gil_release(state);
+    } else
+    { name = "no-GIL";
+    }
 
     SfprintfX(s, "<py_%Us>(%p)", name, obj);
 
@@ -425,24 +454,23 @@ check_error(PyObject *obj)
 
 
 static int py_unify_unicode(term_t t, PyObject *obj, int flags);
+static PyObject *enum_constructor = NULL;
 
 static int
 PyEnum_Check(PyObject *obj)
-{ static PyObject *constructor = NULL;
-
-  if ( !constructor )
+{ if ( !enum_constructor )
   { PyObject *mod_name = NULL;
     PyObject *mod = NULL;
 
     if ( (mod_name=check_error(PyUnicode_FromString("enum"))) &&
          (mod=check_error(PyImport_Import(mod_name))) )
-      constructor = check_error(PyObject_GetAttrString(mod, "Enum"));
+      enum_constructor = check_error(PyObject_GetAttrString(mod, "Enum"));
 
     Py_CLEAR(mod_name);
     Py_CLEAR(mod);
   }
 
-  return PyObject_IsInstance(obj, constructor);
+  return PyObject_IsInstance(obj, enum_constructor);
 }
 
 
@@ -1308,13 +1336,35 @@ py_is_record(PyObject *rec)
 		 *******************************/
 
 #define CVT_TEXT_EX (CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION)
+static int py_thread = 0;
+
+static int
+py_halt(int rc, void *ctx)
+{ (void)ctx;
+  py_finalize();
+
+  return 0;
+}
+
+static void
+py_thread_exitted(void *ctx)
+{ (void) ctx;
+  DEBUG(0, Sdprintf("Thread %d, which created Python has exitted\n",
+		    PL_thread_self()));
+  py_thread = 0;
+}
 
 static int
 py_init(void)
 { if ( !py_initialize_done )
   { predicate_t pred = PL_predicate("py_initialize", 0, "janus");
 
-    return PL_call_predicate(NULL, PL_Q_NORMAL, pred, 0);
+    int rc = PL_call_predicate(NULL, PL_Q_NORMAL, pred, 0);
+    if ( rc )
+    { py_thread = PL_thread_self();
+      PL_thread_at_exit(py_thread_exitted, NULL, FALSE);
+    }
+    return rc;
   }
 
   return TRUE;
@@ -1360,8 +1410,8 @@ py_initialize_(term_t prog, term_t Argv, term_t options)
 #pragma GCC diagnostic ignored "-Wdeprecated-declarations"
 #endif
   Py_SetProgramName(pname);
-  PySys_SetArgv(argc, argv);
   Py_Initialize();
+  PySys_SetArgv(argc, argv);
 #ifndef _MSC_VER
 #pragma GCC diagnostic pop
 #endif
@@ -1377,7 +1427,9 @@ py_initialize_(term_t prog, term_t Argv, term_t options)
 
   PyConfig_Clear(&config);
 #endif
+  py_yield_first();
   py_initialize_done = TRUE;
+  PL_exit_hook(py_halt, NULL);	/* called just before exit */
   rc = TRUE;
   goto succeeded;
 
@@ -1394,6 +1446,7 @@ failed:
   rc = FALSE;
 
 succeeded:
+
   if ( argv )
     free(argv);
   PL_STRINGS_RELEASE();
@@ -1587,16 +1640,40 @@ unchain(term_t call, PyObject **py_target)
   return rc;
 }
 
+
+/* py_gil_ensure()  and py_gil_release()  is  our  wrapper around  the
+ * Python GIL.   The first call  lazily initializes Python  if needed.
+ * This  thread   than  holds   the  GIL.    This  thread   must  call
+ * PyEval_SaveThread() to allow other  threads to interact with Python
+ * and call PyEval_RestoreThread() when it wants to call Python again.
+ * This,  while other  threads  need to  call PyGILState_Ensure()  and
+ * PyGILState_Release() to allow for calling Python.
+ *
+ * I  am not  sure this  is  correct.  Please  comment if  there is  a
+ * simpler way to achieve what we want.
+ *
+ * If we  have the GIL  and we  have delayed Py_DECREF()  call pending
+ * from the atom garbage collector we call delayed_decref();
+ */
+
 static int
-py_gil_ensure(PyGILState_STATE *state)
-{ if ( !py_init() )
+py_gil_ensure(py_gil_state *state)
+{ int self = PL_thread_self();
+
+  if ( !py_init() )
     return FALSE;
 
-  py_resume();
-  *state = PyGILState_Ensure();
+  state->use_gil = (self != py_thread);
+  if ( state->use_gil )
+    state->gil = PyGILState_Ensure();
+  else
+    py_resume(state);
+
+  py_gil_thread = self;
 #ifndef HAVE_PYGILSTATE_CHECK
   have_gil = TRUE;
 #endif
+
   if ( delayed )
     delayed_decref(NULL);
 
@@ -1605,15 +1682,16 @@ py_gil_ensure(PyGILState_STATE *state)
 
 
 static void
-py_gil_release(PyGILState_STATE state)
-{ if ( state )
-  {
+py_gil_release(py_gil_state state)
+{
 #ifndef HAVE_PYGILSTATE_CHECK
-    have_gil = FALSE;
+  have_gil = FALSE;
 #endif
-    PyGILState_Release(state);
-    py_yield();
-  }
+  py_gil_thread = 0;
+  if ( state.use_gil )
+    PyGILState_Release(state.gil);
+  else
+    py_yield(state);
 }
 
 
@@ -1668,7 +1746,7 @@ py_call3(term_t Call, term_t result, term_t options)
   term_t call = PL_copy_term_ref(Call);
   term_t val = 0;
   int rc = TRUE;
-  PyGILState_STATE state;
+  py_gil_state state;
   int uflags = 0;
 
   if ( !get_conversion_options(options, &uflags) )
@@ -1763,7 +1841,7 @@ static foreign_t
 py_iter3(term_t Iterator, term_t Result, term_t options, control_t handle)
 { iter_state iter_buf;
   iter_state *state;
-  PyGILState_STATE gil_state;
+  py_gil_state gil_state;
   int rc = FALSE;
 
   switch( PL_foreign_control(handle) )
@@ -1851,7 +1929,7 @@ py_iter2(term_t Iterator, term_t Result, control_t handle)
 
 static foreign_t
 py_with_gil(term_t goal)
-{ PyGILState_STATE state;
+{ py_gil_state state;
 
   if ( !py_gil_ensure(&state) )
     return FALSE;
@@ -1862,14 +1940,21 @@ py_with_gil(term_t goal)
 }
 
 
+static foreign_t
+py_gil_owner(term_t owner)
+{ if ( py_gil_thread )
+    return PL_unify_thread_id(owner, py_gil_thread);
+
+  return FALSE;
+}
+
+
 static PL_option_t pyrun_options[] =
 { PL_OPTION("file_name", OPT_STRING),
   PL_OPTION("start",     OPT_ATOM),
   PL_OPTIONS_END
 };
 
-
-
 static foreign_t
 py_run(term_t Cmd, term_t Globals, term_t Locals, term_t Result, term_t options)
 { char *cmd;
@@ -1877,7 +1962,7 @@ py_run(term_t Cmd, term_t Globals, term_t Locals, term_t Result, term_t options)
   if ( PL_get_chars(Cmd, &cmd, CVT_ATOM|CVT_STRING|CVT_LIST|CVT_EXCEPTION) )
   { PyObject *locals=NULL, *globals=NULL;
     PyObject *result = NULL;
-    PyGILState_STATE state;
+    py_gil_state state;
     int rc;
     char *file_name = "string";
     atom_t start = ATOM_file;
@@ -1930,7 +2015,7 @@ static foreign_t
 py_str(term_t t, term_t str)
 { PyObject *obj;
   int rc;
-  PyGILState_STATE state;
+  py_gil_state state;
 
   if ( !py_gil_ensure(&state) )
     return FALSE;
@@ -1983,36 +2068,77 @@ py_is_object(term_t t)
 }
 
 
-
 		 /*******************************
 		 *        GIL MANAGEMENT        *
 		 *******************************/
 
 typedef struct
 { PyThreadState *state;
-  int		 yielded;
+  int		 nested;
 } py_state_t;
 
-static _Thread_local py_state_t py_state;
+static py_state_t py_state;
 
 static void
-py_yield(void)
-{ if ( !py_state.yielded && PyGILState_GetThisThreadState() )
+py_yield_first(void)
+{ py_state.state = PyEval_SaveThread();
+  py_state.nested = 0;
+}
+
+static void
+py_yield(py_gil_state state)
+{ if ( --py_state.nested == 0 )
   { DEBUG(1, Sdprintf("Yielding ..."));
     py_state.state = PyEval_SaveThread();
     DEBUG(1, Sdprintf("ok\n"));
-    py_state.yielded = TRUE;
+  } else
+  { PyGILState_Release(state.gil);
   }
 }
 
 static void
-py_resume(void)
-{ if ( py_state.yielded )
+py_resume(py_gil_state *state)
+{ if ( py_state.state )
   { DEBUG(1, Sdprintf("Un yielding ..."));
     PyEval_RestoreThread(py_state.state);
     DEBUG(1, Sdprintf("ok\n"));
-    py_state.yielded = FALSE;
+    py_state.state = NULL;
+  } else
+  { state->gil = PyGILState_Ensure();
   }
+  py_state.nested++;
+}
+
+/* Ideally this would shut down Python such that we can restart it, but
+   that requires getting rid of all Python object references.   These
+   exist both at several places in static variables of this module as
+   in Prolog atoms.   So, for now this is just a cleanup after the
+   Prolog cleanup.
+*/
+
+static foreign_t
+py_finalize(void)
+{ if ( py_initialize_done && !py_finalizing )
+  { py_finalizing = TRUE;
+    if ( py_state.state )
+    { PyEval_RestoreThread(py_state.state);
+      py_state.state = NULL;
+    }
+    py_state.nested = 0;
+
+    Py_CLEAR(enum_constructor);
+    Py_FinalizeEx();
+    py_thread = 0;
+
+    py_initialize_done = FALSE;
+    if ( py_module_table )
+    { py_free_hashmap(py_module_table);
+      py_module_table = NULL;
+    }
+    py_finalizing = FALSE;
+  }
+
+  return TRUE;
 }
 
 
@@ -2058,6 +2184,7 @@ install_janus(void)
         PL_register_foreign_in_module("janus", name, arity, func, flags)
 
   REGISTER("py_initialize_",  3, py_initialize_, 0);
+  REGISTER("py_finalize",     0, py_finalize,    0);
   REGISTER("py_call",         1, py_call1,       0);
   REGISTER("py_call",         2, py_call2,       0);
   REGISTER("py_call",         3, py_call3,       0);
@@ -2067,6 +2194,7 @@ install_janus(void)
   REGISTER("py_free",         1, py_free,        0);
   REGISTER("py_is_object",    1, py_is_object,   0);
   REGISTER("py_with_gil",     1, py_with_gil,    PL_FA_TRANSPARENT);
+  REGISTER("py_gil_owner",    1, py_gil_owner,   0);
   REGISTER("py_str",          2, py_str,         0);
   REGISTER("py_debug",        1, py_debug,       0);
 
@@ -2081,6 +2209,6 @@ uninstall_janus(void)
     py_module_table = NULL;
   }
 #if O_DEBUG
-  Py_FinalizeEx();
+  py_finalize();
 #endif
 }
