@@ -43,6 +43,7 @@
             py_free/1,			% +Obj
 	    py_is_object/1,		% @Term
 	    py_with_gil/1,		% :Goal
+	    py_gil_owner/1,		% -ThreadID
 
             py_func/3,                  % +Module, +Func, -Return
             py_func/4,                  % +Module, +Func, -Return, +Options
@@ -65,12 +66,15 @@
 
             py_initialize/3,            % +Program, +Argv, +Options
             py_lib_dirs/1,              % -Dirs
+            py_add_lib_dir/0,           % (directive)
             py_add_lib_dir/1,           % +Dir
             py_add_lib_dir/2,           % +Dir,+Where
 
             op(200, fy, @),             % @constant
             op(50,  fx, #)              % #Value
           ]).
+:- meta_predicate py_with_gil(0).
+
 :- use_module(library(apply_macros), []).
 :- autoload(library(lists), [append/3, member/2]).
 :- autoload(library(apply), [maplist/2, exclude/3, maplist/3]).
@@ -97,8 +101,6 @@ add_python_dll_dir :-
 :- use_foreign_library(foreign(janus), [visibility(global)]).
 :- endif.
 
-:- meta_predicate py_with_gil(0).
-
 :- predicate_options(py_call/3, 3,
                      [ py_object(boolean),
                        py_string_as(oneof([string,atom]))
@@ -121,10 +123,10 @@ add_python_dll_dir :-
 /** <module> Call Python from Prolog
 
 This library implements calling Python  from   Prolog.  It  is available
-directly from Prolog if the janus   package is bundled, providing access
-to an _embedded_ Python instance. If  SWI-Prolog is embedded into Python
-using the Python package ``janus-swi``, this  library is provided either
-from Prolog or from the Python package.
+directly from Prolog if  the  janus   package  is  bundled.  The library
+provides access to an  _embedded_  Python   instance.  If  SWI-Prolog is
+embedded into Python  using  the   Python  package  ``janus-swi``,  this
+library is provided either from Prolog or from the Python package.
 
 Normally,  the  Prolog  user  can  simply  start  calling  Python  using
 py_call/2 or friends. In special cases it   may  be needed to initialize
@@ -170,15 +172,19 @@ py_version :-
 %       % call a built-in
 %	?- py_call(print("Hello World!\n")).
 %	true.
+%
 %       % call a built-in (alternative)
 %	?- py_call(builtins:print("Hello World!\n")).
 %	true.
+%
 %	% call function in a module
 %	?- py_call(sys:getsizeof([1,2,3]), Size).
 %	Size = 80.
+%
 %	% call function on an attribute of a module
 %       ?- py_call(sys:path:append("/home/bob/janus")).
 %       true
+%
 %       % get attribute from a module
 %       ?- py_call(sys:path, Path)
 %       Path = ["dir1", "dir2", ...]
@@ -203,8 +209,10 @@ py_version :-
 %
 %       ?- py_call(dog:'Dog'("Fido"), Dog).
 %       Dog = <py_Dog>(0x7f095c9d02e0).
+%
 %       ?- py_call($Dog:add_trick("roll_over")).
 %       Dog = <py_Dog>(0x7f095c9d02e0).
+%
 %       ?- py_call($Dog:tricks, Tricks).
 %       Dog = <py_Dog>(0x7f095c9d02e0),
 %       Tricks = ["roll_over"]
@@ -239,8 +247,9 @@ py_version :-
 %       object reference. Some objects are _always_ translated to
 %       Prolog, regardless of this flag.  These are the Python constants
 %       ``None``, ``True`` and ``False`` as well as instances of the
-%       Python base classes long, float, string or tuple.  Instances of
-%       sub classes of these base classes are controlled by this option.
+%       Python base classes `int`, `float`, `str` or `tuple`. Instances
+%       of sub classes of these base classes are controlled by this
+%       option.
 
 %!  py_iter(+Iterator, -Value) is nondet.
 %!  py_iter(+Iterator, -Value, +Options) is nondet.
@@ -269,7 +278,7 @@ py_version :-
 %   Prolog predicate succeeds deterministically. On   success,  the next
 %   candidate is stored.
 %
-%   Note that a Python _generator_  is   a  Python _iterator. Therefore,
+%   Note that a Python _generator_ is   a  Python _iterator_. Therefore,
 %   given  the  Python  generator   expression    below,   we   can  use
 %   py_iter(squares(1,5),X) to generate the squares on backtracking.
 %
@@ -280,7 +289,7 @@ py_version :-
 %   ```
 %
 %   @arg Options is processed as with py_call/3.
-%   @bug Iterator may not depend on janus.Query()
+%   @bug Iterator may not depend on janus.query()
 
 %!  py_run(+String, +Globals, +Locals, -Result, +Options) is det.
 %
@@ -305,12 +314,12 @@ py_version :-
 
 %!  py_free(+Obj) is det.
 %
-%   Immediately free (decrement  the  reference   count)  for  th Python
-%   object Obj. Further reference to Obj  using py_call/1,2 or py_free/1
-%   raises an `existence_error`. Note that by decrementing the reference
-%   count, we make the  reference  invalid   from  Prolog.  This may not
-%   actually delete the object because the   object  may have references
-%   inside Python.
+%   Immediately free (decrement the  reference   count)  for  the Python
+%   object Obj. Further reference  to  Obj   using  e.g.,  py_call/2  or
+%   py_free/1 raises an `existence_error`. Note that by decrementing the
+%   reference count, we make the reference invalid from Prolog. This may
+%   not  actually  delete  the  object  because   the  object  may  have
+%   references inside Python.
 %
 %   Prolog references to Python objects  are   subject  to  atom garbage
 %   collection and thus normally do not need to be freed explicitly.
@@ -318,10 +327,23 @@ py_version :-
 %!  py_with_gil(:Goal) is semidet.
 %
 %   Run Goal as  once(Goal)  while  holding   the  Phyton  GIL  (_Global
-%   Interpreter Lock_). Note that py_call/1,2 also   locks the GIL. This
-%   predicate is only required if we  wish   to  make  multiple calls to
-%   Python while keeping the GIL. The GIL is a _recursive_ lock and thus
-%   calling py_call/1,2 while holding the GIL does not _deadlock_.
+%   Interpreter Lock_). Note that  all   predicates  that  interact with
+%   Python lock the GIL. This predicate is   only required if we wish to
+%   make multiple calls to Python while keeping   the  GIL. The GIL is a
+%   _recursive_ lock and thus calling py_call/1,2  while holding the GIL
+%   does not _deadlock_.
+
+%!  py_gil_owner(-Thread) is semidet.
+%
+%   True when  the Python GIL is  owned by Thread.  Note  that, unless
+%   Thread  is the  calling thread,  this merely  samples the  current
+%   state and may thus no longer  be true when the predicate succeeds.
+%   This predicate is intended to help diagnose _deadlock_ problems.
+%
+%   Note that  this predicate returns  the Prolog threads  that locked
+%   the GIL.  It is however possible that Python releases the GIL, for
+%   example if  it performs a  blocking call.  In this  scenario, some
+%   other thread or no thread may hold the gil.
 
 
 		 /*******************************
@@ -465,7 +487,7 @@ comma_dict_items(Key:Value, Keys) =>
 %
 %       ?- py_shell.
 %       ...
-%       >>> once("writeln(X)", {"X":"Hello world"})
+%       >>> query_once("writeln(X)", {"X":"Hello world"})
 %       Hello world
 %       {'truth': True}
 %
@@ -499,10 +521,13 @@ import_janus :-
 %!  py_pp(+Term, +Options) is det.
 %!  py_pp(+Stream, +Term, +Options) is det.
 %
-%   Pretty prints the Prolog translation of   a Python data structure in
-%   Python syntax. This  exploits  pformat()   from  the  Python  module
-%   `pprint` to do the actual  formatting.   Options  is translated into
-%   keyword arguments passed to pprint.pformat().  For example:
+%   Pretty prints the Prolog translation of a Python data structure in
+%   Python  syntax. This  exploits  pformat() from  the Python  module
+%   `pprint` to do the actual  formatting.  Options is translated into
+%   keyword arguments  passed to  pprint.pformat().  In  addition, the
+%   option  nl(Bool)  is processed.   When  `true`  (default), we  use
+%   pprint.pp(), which  makes the output  followed by a  newline.  For
+%   example:
 %
 %   ```
 %   ?- py_pp(py{a:1, l:[1,2,3], size:1000000},
@@ -517,8 +542,13 @@ py_pp(Term, Options) :-
     py_pp(current_output, Term, Options).
 
 py_pp(Stream, Term, Options) :-
-    opts_kws(Options, Kws),
-    PFormat =.. [pformat, Term|Kws],
+    select_option(nl(NL), Options, Options1, true),
+    (   NL == true
+    ->  Method = pp
+    ;   Method = pformat
+    ),
+    opts_kws(Options1, Kws),
+    PFormat =.. [Method, Term|Kws],
     py_call(pprint:PFormat, String),
     write(Stream, String).
 
@@ -660,38 +690,55 @@ py_lib_dirs(Dirs) :-
     py_call(sys:path, Dirs0),
     maplist(prolog_to_os_filename, Dirs, Dirs0).
 
+%!  py_add_lib_dir is det.
 %!  py_add_lib_dir(+Dir) is det.
 %!  py_add_lib_dir(+Dir, +Where) is det.
 %
 %   Add a directory to the Python  module   search  path.  In the second
 %   form, Where is one of `first`   or `last`. py_add_lib_dir/1 adds the
-%   directory as last.
+%   directory as first. The property `sys:path`   is  not modified if it
+%   already contains Dir.
 %
 %   Dir is in Prolog notation. The added   directory  is converted to an
 %   absolute path using the OS notation.
 %
-%   A flexible way to add the directory  holding the current Prolog file
-%   to the Python search path is  in   the  template below. The `here/0`
-%   predicate can be replaced by  any   predicate  defined  in the file,
-%   either above or below the initializing/1   directive.  A simple name
-%   like here/0 is good style when this code is part of a Prolog module.
-%
-%   ```
-%   here.
-%   :- initialization
-%       (   source_file(here, File),
-%           file_directory_name(File, Dir),
-%           py_add_lib_dir(Dir, first)
-%       ).
-%   ```
+%   The form py_add_lib_dir/0 may only be  used as a _directive_, adding
+%   the directory from which the current Prolog  source is loaded at the
+%   head  of  the   Python   search    path.   If   py_add_lib_dir/1  or
+%   py_add_lib_dir/2 are used in a directive  and the given directory is
+%   not absolute, it is  resolved  against   the  directory  holding the
+%   current Prolog source.
+
+:- multifile system:term_expansion/2.
+
+system:term_expansion((:- py_add_lib_dir),
+                      (:- initialization(py_add_lib_dir(Dir, first), now))) :-
+    prolog_load_context(directory, Dir).
+system:term_expansion((:- py_add_lib_dir(Dir0)),
+                      (:- initialization(py_add_lib_dir(Dir, first), now))) :-
+    \+ is_absolute_file_name(Dir0),
+    prolog_load_context(directory, CWD),
+    absolute_file_name(Dir0, Dir, [relative_to(CWD)]).
+system:term_expansion((:- py_add_lib_dir(Dir0, Where)),
+                      (:- initialization(py_add_lib_dir(Dir, Where), now))) :-
+    \+ is_absolute_file_name(Dir0),
+    prolog_load_context(directory, CWD),
+    absolute_file_name(Dir0, Dir, [relative_to(CWD)]),
+    absolute_file_name(Dir0, Dir).
+
+py_add_lib_dir :-
+    throw(error(context_error(nodirective, py_add_lib_dir), _)).
 
 py_add_lib_dir(Dir) :-
-    py_add_lib_dir(Dir, last).
+    py_add_lib_dir(Dir, first).
 
 py_add_lib_dir(Dir, Where) :-
     absolute_file_name(Dir, AbsDir),
     prolog_to_os_filename(AbsDir, OSDir),
-    (   Where == last
+    (   py_call(sys:path, Dirs0),
+        memberchk(OSDir, Dirs0)
+    ->  true
+    ;   Where == last
     ->  py_call(sys:path:append(OSDir), _)
     ;   Where == first
     ->  py_call(sys:path:insert(0, OSDir), _)
@@ -709,7 +756,7 @@ py_add_lib_dir(Dir, Where) :-
 
 %   py_call_string(:String, +DictIn, -Dict) is nondet.
 %
-%   Support janus.once() and janus.Query(). Parses   String  into a goal
+%   Support janus.query_once() and janus.query(). Parses   String  into a goal
 %   term. Next, all variables from the goal   term that appear in DictIn
 %   are bound to the value from  this   dict.  Dict  is created from the
 %   remaining variables, unless they  start   with  an underscore (e.g.,
@@ -808,15 +855,34 @@ unbind(NonVar, NonVar).
 
 :- public
        px_cmd/3,
-       px_qdet/4,
+       px_call/4,
        px_comp/7.
 
 % These predicates are helpers  for the corresponding Python functions
 % in janus.py.
 
-px_cmd(M, P, Tuple) :-
-    atom_string(Pred, P),
-    atom_string(Module, M),
+
+%   px_call(+Input:tuple, +Module, -Pred, -Ret)
+%
+%   Supports px_qdet() and apply()
+
+px_call(-(), Module, Pred, Ret) =>
+    call(Module:Pred, Ret).
+px_call(-(A1), Module, Pred, Ret) =>
+    call(Module:Pred, A1, Ret).
+px_call(-(A1,A2), Module, Pred, Ret) =>
+    call(Module:Pred, A1, A2, Ret).
+px_call(-(A1,A2,A3), Module, Pred, Ret) =>
+    call(Module:Pred, A1, A2, A3, Ret).
+px_call(-(A1,A2,A3,A4), Module, Pred, Ret) =>
+    call(Module:Pred, A1, A2, A3, A4, Ret).
+px_call(Tuple, Module, Pred, Ret) =>
+    compound_name_arguments(Tuple, _, Args),
+    append(Args, [Ret], GArgs),
+    Goal =.. [Pred|GArgs],
+    call(Module:Goal).
+
+px_cmd(Module, Pred, Tuple) :-
     (   compound(Tuple)
     ->  compound_name_arguments(Tuple, _, Args),
 	Goal =.. [Pred|Args]
@@ -824,20 +890,7 @@ px_cmd(M, P, Tuple) :-
     ),
     call(Module:Goal).
 
-px_qdet(M, P, Tuple, Ret) :-
-    atom_string(Pred, P),
-    atom_string(Module, M),
-    (   compound(Tuple)
-    ->  compound_name_arguments(Tuple, _, Args),
-	append(Args, [Ret], GArgs),
-	Goal =.. [Pred|GArgs],
-	call(Module:Goal)
-    ;   call(Module:Pred, Ret)
-    ).
-
-px_comp(M, P, Tuple, Vars, Set, TV, Ret) :-
-    atom_string(Pred, P),
-    atom_string(Module, M),
+px_comp(Module, Pred, Tuple, Vars, Set, TV, Ret) :-
     length(Out, Vars),
     (   compound(Tuple)
     ->  compound_name_arguments(Tuple, _, Args),
