@@ -122,12 +122,20 @@ swipl_call(PyObject *self, PyObject *args)
   fid_t fid;
   Py_ssize_t arity = PyTuple_GET_SIZE(args);
 
+  if ( py_finalizing )
+    Py_RETURN_NONE;		/* error? */
+
   if ( arity == 0 || arity > 3 )
   { PyErr_SetString(PyExc_TypeError,
 		    "swipl.call(query,bindings,keep) takes 1..3 arguments");
     return NULL;
   }
 
+  if ( PL_thread_attach_engine(NULL) < 0 )
+  { Py_SetPrologErrorFromChars("Cannot create thread");
+    return NULL;
+  }
+
   if ( !pred || !user )
   { pred = PL_predicate("py_call_string", 3, "janus");
     user = PL_new_module(PL_new_atom("user"));
@@ -140,7 +148,12 @@ swipl_call(PyObject *self, PyObject *args)
 	 unify_input(av+1, arity, args) )
     { qid_t qid = PL_open_query(user, PL_Q_CATCH_EXCEPTION|PL_Q_EXT_STATUS,
 				pred, av);
-      int rc = PL_next_solution(qid);
+      int rc;
+
+      Py_BEGIN_ALLOW_THREADS
+      rc = PL_next_solution(qid);
+      Py_END_ALLOW_THREADS
+
       switch(rc)
       { case PL_S_TRUE:
 	case PL_S_LAST:
@@ -168,10 +181,123 @@ swipl_call(PyObject *self, PyObject *args)
     else
       PL_discard_foreign_frame(fid);
   }
+  PL_thread_destroy_engine();
 
   return out;
 }
 
+
+static atom_t
+py_obj_to_atom(PyObject *obj, const char *ex)
+{ if ( PyUnicode_Check(obj) )
+  { ssize_t len;
+    wchar_t *s;
+    atom_t a;
+
+    s = PyUnicode_AsWideCharString(obj, &len);
+    if ( !check_error((void*)s) )
+      return 0;
+    a = PL_new_atom_wchars(len, s);
+    PyMem_Free(s);
+    return a;
+  }
+
+  PyErr_SetString(PyExc_TypeError, ex);
+  return 0;
+}
+
+
+static PyObject *
+swipl_apply_once(PyObject *self, PyObject *args, PyObject *kwargs)
+{ Py_ssize_t argc = PyTuple_GET_SIZE(args);
+  PyObject *rc = NULL;
+  atom_t mname=0;
+  atom_t pname=0;
+
+  if ( argc >= 2 )
+  { fid_t fid;
+    PyObject *on_fail = NULL;
+
+    if ( kwargs )
+    { static const char *kwds[] = {"fail", NULL};
+      static PyObject *empty = NULL;
+
+      if ( !empty && !(empty = PyTuple_New(0)) )
+	return NULL;
+      if ( !PyArg_ParseTupleAndKeywords(empty, kwargs, "|$O",
+					(char**)kwds, &on_fail) )
+	return NULL;
+    }
+
+    if ( !(mname=py_obj_to_atom(PyTuple_GetItem(args, 0),
+				"module expected")) )
+      goto error;
+    if ( !(pname=py_obj_to_atom(PyTuple_GetItem(args, 1),
+				"predicate name expected")) )
+      goto error;
+
+    if ( (fid=PL_open_foreign_frame()) )
+    { term_t av;
+
+      if ( (av=PL_new_term_refs(argc-1)) )
+      { for(Py_ssize_t i=2; i < argc; i++)
+	{ if ( !py_unify(av+i-2, PyTuple_GetItem(args, i), 0) )
+	    goto eunify;
+
+	}
+      }
+      module_t m = PL_new_module(mname);
+      functor_t f = PL_new_functor(pname, argc-1);
+      predicate_t pred = PL_pred(f, m);
+      qid_t qid;
+
+      if ( (qid=PL_open_query(m, PL_Q_CATCH_EXCEPTION|PL_Q_EXT_STATUS,
+			      pred, av)) )
+      { int r;
+
+	Py_BEGIN_ALLOW_THREADS
+	r = PL_next_solution(qid);
+	Py_END_ALLOW_THREADS
+
+	switch(r)
+	{ case PL_S_TRUE:
+	  case PL_S_LAST:
+	    PL_cut_query(qid);
+	    if ( !py_from_prolog(av+argc-2, &rc) )
+	      Py_SetPrologError(PL_exception(0));
+	    break;
+	  case PL_S_EXCEPTION:
+	    Py_SetPrologError(PL_exception(qid));
+	    PL_cut_query(qid);
+	    break;
+	  case PL_S_FALSE:
+	    PL_cut_query(qid);
+	    if ( on_fail )
+	    { rc = on_fail;
+	      Py_INCREF(rc);
+	    } else
+	      Py_SetPrologErrorFromChars("apply_once(): goal failed");
+	    break;
+	  default:
+	    assert(0);
+	}
+      }
+
+    eunify:
+      PL_discard_foreign_frame(fid);
+    }
+  } else
+  { PyErr_SetString(PyExc_TypeError, "swipl.apply_once(module, predicate, [input ...]) expected");
+  }
+
+error:
+  if ( mname ) PL_unregister_atom(mname);
+  if ( pname ) PL_unregister_atom(pname);
+
+  return rc;
+}
+
+
 static void
 tuple_set_int(int i, PyObject *tuple, int64_t val)
 { PyObject *v = PyLong_FromLongLong(val);
@@ -192,6 +318,11 @@ swipl_open_query(PyObject *self, PyObject *args)
     return NULL;
   }
 
+  if ( PL_thread_attach_engine(NULL) < 0 )
+  { Py_SetPrologErrorFromChars("Cannot create thread");
+    return NULL;
+  }
+
   if ( !pred || !user )
   { pred = PL_predicate("py_call_string", 3, "janus");
     user = PL_new_module(PL_new_atom("user"));
@@ -214,6 +345,7 @@ swipl_open_query(PyObject *self, PyObject *args)
     }
   }
 
+  PL_thread_destroy_engine();
   Py_SetPrologError(PL_exception(0));
   return NULL;
 }
@@ -266,13 +398,16 @@ swipl_next_solution(PyObject *self, PyObject *args)
   int done = FALSE;
   PyObject *tp;
   int keep;
+  int rc;
 
   if ( !query_parms(args, &tp, &fid, &qid, &av, &keep) )
     return NULL;
   if ( !qid )
     return PyBool_FromLong(0);
 
-  int rc = PL_next_solution(qid);
+  Py_BEGIN_ALLOW_THREADS
+  rc = PL_next_solution(qid);
+  Py_END_ALLOW_THREADS
   PyObject *out = NULL;
 
   switch(rc)
@@ -303,7 +438,9 @@ swipl_next_solution(PyObject *self, PyObject *args)
       PL_close_foreign_frame(fid);
     else
       PL_discard_foreign_frame(fid);
+
     tuple_set_int(1, tp, 0);   /* set qid to 0 */
+    PL_thread_destroy_engine();
   }
 
   return out;
@@ -329,6 +466,8 @@ swipl_close_query(PyObject *self, PyObject *args)
       PL_close_foreign_frame(fid);
     else
       PL_discard_foreign_frame(fid);
+
+    PL_thread_destroy_engine();
     tuple_set_int(1, tp, 0);   /* set qid to 0 */
   }
 
@@ -336,6 +475,33 @@ swipl_close_query(PyObject *self, PyObject *args)
 }
 
 
+static PyObject *
+swipl_engine(PyObject *self, PyObject *args)
+{ int tid = PL_thread_self();
+
+  return PyLong_FromLongLong(tid);
+}
+
+static PyObject *
+swipl_attach_engine(PyObject *self, PyObject *args)
+{ int tid = PL_thread_attach_engine(NULL);
+  if ( tid >= 0 )
+    return PyLong_FromLongLong(tid);
+
+  Py_SetPrologErrorFromChars("Cannot create thread");
+  return NULL;
+}
+
+static PyObject *
+swipl_detach_engine(PyObject *self, PyObject *args)
+{ if ( PL_thread_destroy_engine() )
+    Py_RETURN_NONE;
+
+  Py_SetPrologErrorFromChars("No thread to detach");
+  return NULL;
+}
+
+
 static PyObject *
 swipl_erase(PyObject *self, PyObject *args)
 { PyObject *rec = NULL;
@@ -409,12 +575,22 @@ swipl_initialize(PyObject *self, PyObject *args)
 static PyMethodDef swiplMethods[] =
 { {"call", swipl_call, METH_VARARGS,
    "Execute a Prolog query."},
+  {"apply_once", (PyCFunction)swipl_apply_once, METH_VARARGS|METH_KEYWORDS,
+   "Evaluate predicate as function.\n\n"
+   "Synopsis: apply_once(module, predicate, input ...) -> output"
+  },
   {"open_query", swipl_open_query, METH_VARARGS,
    "Open a Prolog query."},
   {"next_solution", swipl_next_solution, METH_VARARGS,
    "Compute the next answer."},
   {"close_query", swipl_close_query, METH_VARARGS,
    "Close an open query."},
+  {"engine", swipl_engine, METH_VARARGS,
+   "Return the engine id of the attached Prolog engine."},
+  {"attach_engine", swipl_attach_engine, METH_VARARGS,
+   "Attach a Prolog engine to the current thread."},
+  {"detach_engine", swipl_detach_engine, METH_VARARGS,
+   "Detach the engine from the current thread."},
   {"erase", swipl_erase, METH_VARARGS,
    "Erase a record."},
 #ifdef PYTHON_PACKAGE
