@@ -63,11 +63,7 @@ test_janus :-
 		xsb_call
               ]).
 
-:- initialization
-    source_file(test_janus, File),
-    file_directory_name(File, Dir0),
-    directory_file_path(Dir0, tests, Dir),
-    py_add_lib_dir(Dir, first).
+:- py_add_lib_dir(tests).
 
 :- begin_tests(janus_data).
 
@@ -287,13 +283,13 @@ test(free, error(existence_error('PyObject', Obj))) :-
 :- begin_tests(python_call_prolog).
 
 test(iter, Sum == 500500) :-
-    py_call(demo:bench_iter(1000), Sum).
+    py_call(demo:bench_query_iter(1000), Sum).
 test(iter, Sum == 10011) :-
     py_call(demo:abort_iter(1000), Sum).
 test(iter, Sum == 10011) :-             % Check we didn't mess up the Prolog stack
     py_call(demo:abort_iter(1000), Sum).
 test(undef, Result == py{truth:Undef}) :-
-    py_call(janus:once(undefined), Result),
+    py_call(janus:query_once(undefined), Result),
     py_call(janus:undefined, Undef).
 test(russel, List == [py{'X':barber,'Y':barber,truth:Undef},
 		      py{'X':barber,'Y':mayor,truth: @true}]) :-
@@ -311,6 +307,20 @@ test(invalid_nesting, X == @true) :-
 test(while, X == [1,2,3]) :-
     py_call(while:test_while(), X).
 :- endif.
+test(apply, X == 3) :-
+    py_call(janus:apply_once(user, plus, 1, 2), X).
+test(apply, X == 1) :-
+    py_call(janus:apply_once(user, between, 1, 2), X).
+test(apply, X == 0) :-
+    py_call(janus:apply_once(user, between, 3, 2, fail=0), X).
+test(apply, error(python_error('PrologError',_,_))) :-
+    py_call(janus:apply_once(user, between, 3, 2), _).
+test(apply, error(python_error('PrologError',_,_))) :-
+    py_call(janus:apply_once(janus, no_such_predicate), _).
+test(apply, X == 1) :-
+    py_call(janus:apply_once(user, =, 1), X).
+test(apply, error(python_error('PrologError',_,_))) :-
+    py_call(janus:apply_once(user, =, prolog(f(a))), _).
 
 :- end_tests(python_call_prolog).
 
