@@ -68,15 +68,9 @@ def kwd_all(a=1, b=2, c=3):
 
 # Demo for dealing with iterators
 
-def bench_iter(n):
-    sum=0
-    for d in Query("between(1,M,X)", {"M":n}):
-        sum += d["X"]
-    return sum;
-
 def abort_iter(n):
     sum=0
-    for d in Query("between(1,M,X)", {"M":n}):
+    for d in query("between(1,M,X)", {"M":n}):
         sum += d["X"]
         if sum > 10000:
             break;
@@ -84,14 +78,14 @@ def abort_iter(n):
 
 def double_iter(w,h):
     tuples=[]
-    for yd in Query("between(1,M,Y)", {"M":h}):
-        for xd in Query("between(1,M,X)", {"M":w}):
+    for yd in query("between(1,M,Y)", {"M":h}):
+        for xd in query("between(1,M,X)", {"M":w}):
             tuples.append((xd['X'],yd['Y']))
     return tuples
 
 def test_invalid_nesting():
-    q1 = Query("between(1,3,X)")
-    q2 = Query("between(1,3,Y)")
+    q1 = query("between(1,3,X)")
+    q2 = query("between(1,3,Y)")
     q2.next()
     try:
         q1.next()
@@ -110,17 +104,43 @@ def squares(start, stop):
 # Test for WFS
 
 def shaves(truth=PLAIN_TRUTHVALS):
-    return [*Query("russel:shaves(X,Y)", truth=truth)]
+    return [*query("russel:shaves(X,Y)", truth=truth)]
 
 # Benchmarking support
 
+def bench_query_iter(n):
+    sum=0
+    for d in query("between(1,M,X)", {"M":n}):
+        sum += d["X"]
+    return sum;
+
+def bench_apply_iter(n):
+    sum=0
+    for d in apply("user", "between", 1, n):
+        sum += d
+    return sum;
+
 def bench_call(n):
     for i in range(1,n):
-        once("Y is X+1", {"X":i})
+        query_once("Y is X+1", {"X":i})
 
 def bench_call_v(n):
     for i in range(1,n):
-        print((i, once("py_thread(T)", {})["T"]))
+        print((i, query_once("py_thread(T)", {})["T"]))
+
+def bench_apply_once(n):
+    sum=0
+    for i in range(1,n+1):
+        sum = sum + apply_once("user", "=", i)
+    return sum
+
+def bench_apply_oncea(n):
+    for i in range(1,n+1):
+        apply_once("user", "between", 1, 2)
+
+def bench_apply_onceb(n):
+    for i in range(1,n+1):
+        apply_once("user", "between", 1, 2, fail=0)
 
 def bench_px_cmd(n):
     """Call `true` `n` times"""
@@ -151,7 +171,7 @@ class GCAble:
 def call_except(query, input={}):
     """Call query, return the exception or none"""
     try:
-        once(query, input)
+        query_once(query, input)
     except Exception as ex:
         return ex
     return None
@@ -161,7 +181,7 @@ def call_try(n, query, input={}):
     count = 0
     for i in range(0,n):
         try:
-            once(query, input)
+            query_once(query, input)
         except Exception as ex:
             count = count+1
     return count
