@@ -108,6 +108,7 @@
 #define MSG_DETERMINISM		 62
 
 #define MSG_READ_TOKEN		 65
+#define MSG_LOCALE		 66
 
 #define MSG_CONTINUE		 70
 
@@ -179,6 +180,7 @@
 #define MSG_RECONSULT_PRED	 191
 #define MSG_RECONSULT_CLAUSE	 192
 #define MSG_RECONSULT_MODULE	 193
+#define MSG_AUTOLOAD		 194
 
 #define MSG_TRIE_PUT_TERM	 200
 #define MSG_TRIE_GC		 201
@@ -221,6 +223,8 @@
 #define TABLING_NO_SIMPLIFY	 321		/* these affect behaviour */
 #define TABLING_NO_AC		 322
 
+#define MSG_GVAR_LAZY		 330
+
 #define MSG_THREAD_WAIT		 350
 #define MSG_THREAD_SIGNAL	 351
 
