@@ -351,7 +351,7 @@ acquireSourceFileNo(int index)
 { SourceFile sf;
 
   if ( (sf = indexToSourceFile(index)) )
-    return acquireSourceFile(sf);
+    acquireSourceFile(sf);
 }
 
 
@@ -369,7 +369,7 @@ releaseSourceFile(SourceFile sf)
 		 sf->references-1));
 
   if ( sf->references <= 0 )
-  { Sdprintf("Oops: %d references for sourc file %s\n", PL_atom_chars(sf->name));
+  { Sdprintf("Oops: %d references for source file %s\n", sf->references, PL_atom_chars(sf->name));
     sf->references = 0x4000000;
   }
   if ( ATOMIC_DEC(&sf->references) == 0 )
@@ -614,6 +614,7 @@ PRED_IMPL("$source_file_predicates", 2, source_file_predicates, 0)
     { term_t tail = PL_copy_term_ref(A2);
       term_t head = PL_new_term_ref();
       ListCell cell;
+      rc = TRUE;
 
       LOCKSRCFILE(sf);
       for(cell=sf->procedures; rc && cell; cell = cell->next )
@@ -1076,7 +1077,7 @@ reloadContext(DECL_LD SourceFile sf, Procedure proc)
 	  Sdprintf("%s %s ...\n",
 		   true(reload, P_NEW)        ? "New"   :
 		   true(reload, P_NO_CLAUSES) ? "Alien" :
-					        "Reload",
+						"Reload",
 		   predicateName(def)));
   }
 
@@ -1124,7 +1125,8 @@ assertProcedureSource(DECL_LD SourceFile sf, Procedure proc, Clause clause)
       for(cref2 = cref->next; cref2; cref2 = cref2->next)
       { Clause c2 = cref2->value.clause;
 
-	if ( !GLOBALLY_VISIBLE_CLAUSE(c2, reload->generation) )
+	if ( !GLOBALLY_VISIBLE_CLAUSE(c2, reload->generation) ||
+	     true(c2, CL_ERASED) )
 	  continue;
 	if ( true(def, P_MULTIFILE) && c2->owner_no != sf->index )
 	  continue;
@@ -1366,11 +1368,11 @@ fix_metapredicate(p_reload *r)
       clear(def, P_TRANSPARENT);
       set(def, r->flags&P_TRANSPARENT);
 
-      freeCodesDefinition(def, FALSE);
+      freeCodesDefinition(def, TRUE);
     }
   } else if ( true(r, P_META) )
   { setMetapredicateMask(def, r->args);
-    freeCodesDefinition(def, FALSE);
+    freeCodesDefinition(def, TRUE);
   } else if ( true(r, P_TRANSPARENT) )
   { set(def, P_TRANSPARENT);
   }
@@ -1949,7 +1951,7 @@ BeginPredDefs(srcfile)
   PRED_DEF("$start_consult",		2, start_consult,	     0)
   PRED_DEF("$end_consult",		1, end_consult,		     0)
   PRED_DEF("$fixup_reconsult",		1, fixup_reconsult,          0)
-  PRED_DEF("$set_source_files",	        1, set_source_files,	     0)
+  PRED_DEF("$set_source_files",		1, set_source_files,	     0)
   PRED_DEF("$flush_predicate",		1, flush_predicate,          0)
   PRED_DEF("$flushed_predicate",	1, flushed_predicate,	     0)
 EndPredDefs
